from scipy.special import xlogy, xlog1py

def get_loss(L, alpha, beta, penalty=0):
    """Calculate loss
        if penalty is 0 then results will be -2*LogLik

    Args:
        L: list with (n_positive, n_negative) counts
        alpha (float):  pseudo count
        beta (float): beta value to add to denominator
        penalty (float, optional): penalty value in loss function (c). Defaults to 0.

    Returns:
        float: loss
    """
    res = 0.0
    for (nm, nu) in L:
        p = (nm + alpha)/(nm + nu + alpha + beta)
        res += xlogy(nm, p) + xlog1py(nu, -p)
    return -2*res + len(L)*penalty

def get_betas(alpha, M, U):
    """get array with beta values for each fold

    Args:
        alpha (float): pseudo count
        M (int array): Number of mutated sites in each fold
        U (int array): Number of unmutated sites in each fold

    Returns:
        float array: array with beta value for each fold
    """
    my=M/(M+U)
    betas = (alpha*(1.0-my))/my
    return betas




from kmerpapa.pattern_utils import *
import sys
from numba import njit
import math
import numpy as np
from scipy.special import xlogy, xlog1py

def get_right(super_pat, left):
    right = ''
    for i in range(len(super_pat)):
        if super_pat[i] == left[i]:
            right += super_pat[i]
        else:
            right += minus_set[super_pat[i]][left[i]]
    return right

def backtrack(pattern, backtrack_mem, PE):
    pat_num = PE.pattern2num(pattern)
    left_num = backtrack_mem[pat_num]
    if left_num == pat_num:
        return [pattern]
    left_pat = PE.num2pattern(left_num)
    right_pat = get_right(pattern, left_pat)
    return backtrack(left_pat, backtrack_mem, PE) + backtrack(right_pat, backtrack_mem, PE)

def score(M, U):
    p = (M + alpha)/(M + U + alpha + beta)
    res = -2*(xlogy(M, p) + xlog1py(U,-p)) + penalty
    return res

@njit
def handle_pattern(pattern, score_mem, backtrack_mem, M_mem, U_mem):
    pat_num = 0
    for i in range(gen_pat_len):
        pat_num += perm_code_no_np[gpo[i]][pattern[i]] * cgppl[i]
    first = True
    for i in range(gen_pat_len):
        if has_complement[pattern[i]]:
            pcn_row = perm_code_no_np[gpo[i]]
            pat_num_diff  = pat_num - pcn_row[pattern[i]] * cgppl[i]
            for j in range(n_complements[pattern[i]]):
                c1 = complements_tab_1[pattern[i]][j]
                c2 = complements_tab_2[pattern[i]][j]
                pat1_num = pat_num_diff + pcn_row[c1] * cgppl[i]
                pat2_num = pat_num_diff + pcn_row[c2] * cgppl[i]
                new_score = score_mem[pat1_num]+score_mem[pat2_num] 
                if new_score < score_mem[pat_num]:
                    score_mem[pat_num] = new_score
                    backtrack_mem[pat_num] = pat1_num
                if first:
                    M_mem[pat_num] = M_mem[pat1_num] + M_mem[pat2_num]
                    U_mem[pat_num] = U_mem[pat1_num] + U_mem[pat2_num]
                    first = False
    M = M_mem[pat_num]
    U = U_mem[pat_num]
    p = (M + alpha)/(M + U + alpha + beta)
    s = penalty
    if M > 0:
        s += -2.0 * M*math.log(p)
    if U > 0:
        s += -2.0 * U*math.log(1.0-p)
    if s < score_mem[pat_num]:
        score_mem[pat_num] = s
        backtrack_mem[pat_num] = pat_num


def pattern_partition_bottom_up(gen_pat, contextD, alpha_, beta_, penalty_, args, nmut, nunmut, index_mut=0):
    global cgppl
    global gppl
    global gpo
    global gen_pat_len
    global has_complement
    global alpha
    global beta
    global penalty
    alpha = alpha_
    beta = beta_
    penalty = penalty_
    ftype = np.float32
    npat = pattern_max(gen_pat)
    score_mem = np.full(npat, 1e100, dtype=ftype)
    if nmut+nunmut > np.iinfo(np.uint32).max:
        itype = np.uint64
    else:
        itype = np.uint32
    U_mem = np.empty(npat, dtype=itype)
    M_mem = np.empty(npat, dtype=itype)
    backtrack_mem = np.empty(npat, dtype=np.uint64)

    PE = PatternEnumeration(gen_pat)

    cgppl = np.array(get_cum_genpat_pos_level(gen_pat), dtype=np.uint32)
    gppl = np.array(get_genpat_pos_level(gen_pat), dtype=np.uint32)
    gpo = np.array([ord(x) for x in gen_pat], dtype=np.uint32)
    gen_pat_level = pattern_level(gen_pat)
    gpot = tuple(ord(x) for x in gen_pat)

    gen_pat_len = len(gen_pat)

    has_complement = np.full(100,True)
    has_complement[ord('A')] = False
    has_complement[ord('C')] = False
    has_complement[ord('G')] = False
    has_complement[ord('T')] = False

    for pattern in subpatterns_level(gen_pat, 0):
        pat_num = PE.pattern2num(pattern)
        tup = contextD[pattern]
        nm = tup[index_mut]
        nu = tup[-1]
        M_mem[pat_num] = nm
        U_mem[pat_num] = nu
        score_mem[pat_num] = score(nm, nu)
        backtrack_mem[pat_num] = pat_num

    for level in range(1, gen_pat_level+1):
        if args.verbosity > 1:
            print(f'level {level} of {gen_pat_level}', file=sys.stderr)
        for pattern in subpatterns_level_ord_np(gpot, gen_pat_level, level):
            handle_pattern(pattern, score_mem, backtrack_mem, M_mem, U_mem)
    
    pat_num = PE.pattern2num(gen_pat)
    names = backtrack(gen_pat, backtrack_mem, PE)
    return score_mem[pat_num], M_mem[pat_num], U_mem[pat_num], names





from kmerpapa.pattern_utils import *
from kmerpapa.CV_tools import make_all_folds_contextD_kmers
from kmerpapa.score_utils import get_betas
import sys
import numpy
from scipy.special import xlogy, xlog1py

def test_folds(trainM, trainU, testM, testU, alphas, betas):
    '''
    likelihood of test_data on training set
    '''
    p = (trainM + alphas)/(trainM + trainU + alphas + betas)
    return -2 *(xlogy(testM, p) + xlog1py(testU, -p))

def all_kmers(gen_pat, contextD, alphas, args, nmut, nunmut, index_mut=0):
    nf = args.nfolds
    nit = args.iterations
    npat = generality(gen_pat)
    U_mem = numpy.zeros((npat,nf), dtype=numpy.uint64)
    M_mem = numpy.zeros((npat,nf), dtype=numpy.uint64)
    M_sum_test = numpy.zeros(nf)
    U_sum_test = numpy.zeros(nf)
    test_loss = {a_i:[] for a_i in range(len(alphas))}
    train_loss = {a_i:[] for a_i in range(len(alphas))}
    prng = np.random.RandomState(args.seed)
    for iteration in range(nit):
        make_all_folds_contextD_kmers(contextD, U_mem, M_mem, gen_pat, prng)
        M_sum_test = M_mem.sum(axis=0) # n_mut for each fold
        U_sum_test = U_mem.sum(axis=0) # n_unmut for each fold 
        M_sum_train = sum(M_sum_test) - M_sum_test
        U_sum_train = sum(U_sum_test) - U_sum_test

        for a_i in range(len(alphas)):
            alpha = alphas[a_i]
            betas = get_betas(alpha, M_sum_train, U_sum_train)
            sum_test = numpy.zeros(nf)
            sum_train = numpy.zeros(nf)
            for i, context in enumerate(matches(gen_pat)):
                M_test = M_mem[i]
                U_test = U_mem[i]
                M_train = sum(M_test) - M_test
                U_train = sum(U_test) - U_test
                sum_train += test_folds(M_train, U_train, M_train, U_train, alpha, betas)
                sum_test += test_folds(M_train, U_train, M_test, U_test, alpha, betas)
            train_loss[a_i].extend(list(sum_train))
            test_loss[a_i].extend(list(sum_test))

    best_test_loss = 1e100
    best_alpha = None

    for a_i in range(len(alphas)):
        alpha = alphas[a_i]
        test = sum(test_loss[a_i])/nit
        if args.verbosity > 0:
            print(f'alpha={alpha} test_loss={test}', file = sys.stderr)
        if test < best_test_loss:
            best_alpha = alpha
            best_test_loss = test

    return best_alpha, best_test_loss

            
      



nucleotides = ['A','C','G','T']

def read_joint_kmer_counts(f, super_pattern, n_scale=1):
    """Read kmer counts from a file with three columns: kmer, count_mutated count_background

    Args:
        f (file object): input file
        super_pattern: General pattern
        n_scale (int, optional): Option to scale the background counts. Defaults to 1.

    Returns:
        tuple of: 
            dictionary with kmer counts
            total number of unmutated counts
            total number of mutated counts
    """
    contextD = {}
    n_sites = 0
    n_mut = 0
    for line in f:
        kmer, count_mut, count_denominator = line.split()

        if not all(n in nucleotides for n in kmer):
            continue

        try:
            count_denominator = int(count_denominator)
        except:
            count_denominator = int(float(count_denominator))
        try:
            count_mut = int(count_mut)
        except:
            count_mut = int(float(count_mut))

        assert n_scale*count_denominator - count_mut >= 0, f'''
            background counts should be larger than the positive counts
            so that a negative set can be created by subtraction the positive count
            from the background count. Problematic kmer: {kmer}'''

        if not super_pattern is None:
            if kmer not in super_pattern:
                continue
        n_sites += n_scale*count_denominator
        n_mut += count_mut
        contextD[kmer] = (count_mut, n_scale*count_denominator - count_mut)
    f.close()
    return contextD, n_sites-n_mut, n_mut


def downsize_contextD(D, general_pattern, length):
    """Change a dictionary of kmer counts to a dictinary with kmer counts
    for a smaller value of k.

    Args:
        D (dict): kmer count dictonary
        general_patttern (str): the general pattern
        length (int): The new value of k

    Returns:
        tuple with downsized dictionary and downsized general pattern
    """
    res = {}
    start = None
    end = None
    for context in D:
        if start is None:
            assert not length is None
            assert len(context) > length, f'k-mer:{context} cannot be reduced to length {length}'
            radius1 = (len(context)//2)
            radius2 = length//2
            start = radius1-radius2
            end = (radius1-radius2)+length
        counts = D[context]
        context = context[start:end]
        if context not in res:
            res[context] = [0]*len(counts)
        for i in range(len(counts)):
            res[context][i] += counts[i]
    return res, general_pattern[start:end]


def read_dict(f, super_pattern, length=None):
    """Read kmer counts from a file with three columns: kmer, count

    Args:
        f (file object): input file
        super_pattern: General pattern
        length (int, optional): downscale kmers to this length


    Returns:
        tuple of: 
            dictionary with kmer counts
            total number of counts
    """
    D = {}
    all_counts = 0
    start = None
    end = None
    for line in f:
        context, count = line.split()

        if not all(n in nucleotides for n in context):
            continue

        try:
            count = int(count)
        except:
            count = int(float(count))
        assert count>=0, f'negative counts are not allowed, bad line:\n{line.strip()}'

        if start is None:
            if not length is None and length!=len(context):
                assert len(context) > length
                radius1 = (len(context)//2)
                radius2 = length//2
                start = radius1-radius2
                end = (radius1-radius2)+length
            else:
                start = 0
                end = len(context)
        context = context[start:end]


        if not super_pattern is None:
            assert len(super_pattern) == len(context)
            if context not in super_pattern:
                continue
        
        all_counts += count
        if context not in D:
            D[context] = 0
        D[context] += count
    return D, all_counts


def read_postive_and_other(fpos, fother, super_pattern, n_scale=1, background=True):
    """Read input data from two input files

    Args:
        fpos (file obj): Input file with columns: kmer, count
        fother (file obj): Input file with columns: kmer, count
        super_pattern (str): General Pattern
        n_scale (int, optional): Option to scale the background counts. Defaults to 1.
        background (bool, optional): Is the other dictionary based on background counts or
            positive counts? Defaults to True.

    Returns:
        tuple of: 
            dictionary with kmer counts
            total number of unmutated counts
            total number of mutated counts
    """
    posD, allpos = read_dict(fpos, super_pattern)

    otherD, allother = read_dict(fother, super_pattern, length=len(next(iter(posD.keys()))))
    all_contexts = list(set([*posD.keys(),*otherD.keys()]))
    resD = {}
    for context in all_contexts:
        if context in posD:
            count_mut = posD[context]
        else:
            count_mut = 0

        if context in otherD:
            count_denominator = n_scale * otherD[context]
        else:
            count_denominator  = 0

        if background:
            assert count_denominator >= count_mut, '''
                background counts should be larger than the positive counts
                so that a negative set can be created by subtraction the positive count
                from the background count. Problematic k-mer: {context}
                '''
            count_denominator = count_denominator - count_mut

        resD[context] = (count_mut, count_denominator)
    if background:
        allother = allother - allpos

    return resD, allother, allpos


def read_input(args, super_pattern):
    """Read input file and get dictionary with kmer counts

    Args:
        args: Parsed command line arguments
        super_pattern (str): General Pattern

    Returns:
        tuple of: 
            dictionary with kmer counts
            total number of unmutated counts
            total number of mutated counts
    """
    
    assert (args.positive is None) != (args.joint_context_counts is None), '''
        Either the --positive option or the --join_context_counts option (but not both)
        must be used to provide input data.
        '''
    if not args.positive is None:
        assert (args.negative is None) != (args.background is None), '''
            If the --joint_context_counts option is not used then either the --negative or the
            --background option (but not both) must be used.
            '''
        if not args.negative is None:
            contextD, n_unmut, n_mut = read_postive_and_other(args.positive, args.negative, super_pattern, n_scale = 1, background=False)
        else:
            contextD, n_unmut, n_mut = read_postive_and_other(args.positive, args.background, super_pattern, n_scale = 1, background=True)
    else:
            contextD, n_unmut, n_mut = read_joint_kmer_counts(args.joint_context_counts, super_pattern, n_scale = 1)
        
    return contextD, n_unmut, n_mut



import numpy as np
from numba import njit

code = {'A':['A'],
        'C':['C'],
        'G':['G'],
        'T':['T'],
        'R':['A', 'G'],
        'Y':['C', 'T'],
        'S':['G', 'C'],
        'W':['A', 'T'],
        'K':['G', 'T'],
        'M':['A', 'C'],
        'B':['C', 'G', 'T'],
        'D':['A', 'G', 'T'],
        'H':['A', 'C', 'T'],
        'V':['A', 'C', 'G'],
        'N':['A', 'C', 'G', 'T']}

# dictionary from IUPAC character to nucleotide set
set_code = {}
for x in code:
    set_code[x] = frozenset(code[x])

# dictionary from nucleotide set to IUPAC character
inv_code = {}
for key, val in code.items():
    inv_code[frozenset(val)] = key

# minus_set[nucleotide set][nucleotide set] = IUPAC character
minus_set = {}
for x in code:
    minus_set[x] = {}
    for y in code:
        minus = frozenset(set(code[x])-set(code[y]))
        if len(minus)>0:
            minus_set[x][y] = inv_code[minus]


# dictionary from IUPAC character to list of tuples with all two-partitions written with IUPAC characters
complements = {}
for x in code:
    if len(code[x]) == 2:
        complements[x] = [code[x]]

complements['V'] = [('A', 'S'), ('C', 'R'), ('G', 'M')]
complements['H'] = [('A', 'Y'), ('C', 'W'), ('T', 'M')]
complements['D'] = [('A', 'K'), ('G', 'W'), ('T', 'R')]
complements['B'] = [('C', 'K'), ('G', 'Y'), ('T', 'S')]
complements['N'] = [('S','W'),('K','M'),('R','Y'), ('A','B'),('C','D'),('G','H'),('T','V')]


# complements dictionary but with the output tuples IUPAC characters are encoded as ord(x)
complements_ord = {}
for x in complements:
    complements_ord[x] = [(ord(x),ord(y)) for x,y in complements[x]]

# complements dictionary but with the input and output IUPAC characters are encoded as ord(x)
complements_ord2ord = [()]*100

# n_complements[ord(x)] is number of complements of IUPAC character x
n_complements  = np.zeros(90, dtype=int)

# (complements_tab_1[ord(x)][i], complements_tab_1[ord(x)][i]) are the i'th complements pair of the IUPAC character x 
complements_tab_1 = np.zeros((90,7), dtype=int)
complements_tab_2 = np.zeros((90,7), dtype=int)

for x in complements:
    complements_ord2ord[ord(x)] = tuple(complements_ord[x])
    n_complements[ord(x)] = len(complements_ord[x])
    for i in range(len(complements_ord[x])):
        c1,c2 = complements_ord[x][i]
        complements_tab_1[ord(x)][i] = c1
        complements_tab_2[ord(x)][i] = c2


complements_ord2ord = tuple(complements_ord2ord)

perm_code = {'A':['A'],
            'C':['C'],
            'G':['G'],
            'T':['T'],
            'R':['A', 'G', 'R'],
            'Y':['C', 'T', 'Y'],
            'S':['G', 'C', 'S'],
            'W':['A', 'T', 'W'],
            'K':['G', 'T', 'K'],
            'M':['A', 'C', 'M'],
            'B':['C', 'G', 'T', 'S', 'Y', 'K', 'B'],
            'D':['A', 'G', 'T', 'R', 'W', 'K', 'D'],
            'H':['A', 'C', 'T', 'M', 'W', 'Y', 'H'],
            'V':['A', 'C', 'G', 'M', 'R', 'S', 'V'],
            'N':['A', 'C', 'G', 'T', 'R', 'Y', 'S', 'W', 'K', 'M', 'B', 'D', 'H', 'V', 'N']}


# dictionary from IUPAC character to nucleotide set
set_perm_code = {}
for x in perm_code:
    set_perm_code[x] = frozenset(perm_code[x])


# x,y are IUPAC patterns
# y is sub pattern of x
# perm_code_no[x][y] is the number of y in the perm_code list for x
perm_code_no = {}
for x in perm_code:
    perm_code_no[x] = {}
    for i in range(len(perm_code[x])):
        perm_code_no[x][perm_code[x][i]] = i

# same as perm_code_no just in np.array where x and y are encoded as ord(X) and ord(Y)
perm_code_no_np = np.full((90,90), -100, dtype=int)
for x in perm_code:
    for i in range(len(perm_code[x])):
        perm_code_no_np[ord(x),ord(perm_code[x][i])] = i

# perm_code_max[i][x] is the subset of perm_code x where the cardinality is at most i
perm_code_max = [dict() for x in range(4)]
for i in range(4):
    for x in perm_code:
        perm_code_max[i][x] = []
        for y in perm_code[x]:
            if len(code[y])-1 <= i:
                perm_code_max[i][x].append(y)

# perm_code_min_max[i][j][x] is the subset of perm_code x where the cardinality is at least i and at most j
perm_code_min_max = [[dict() for x in range(4)] for y in range(4)]
for i in range(4):
    for j in range(4):
        for x in perm_code:
            perm_code_min_max[i][j][x] = []
            for y in perm_code[x]:
                if len(code[y])-1 >= i and len(code[y])-1 <= j:
                    perm_code_min_max[i][j][x].append(y)

# same as perm_code_min_max[i][j][x] but x is encoded as ord(x)
perm_code_min_max_ord = [[dict() for x in range(4)] for y in range(4)]
for i in range(4):
    for j in range(4):
        for x in perm_code:
            perm_code_min_max_ord[i][j][x] = []
            for y in perm_code[x]:
                if len(code[y])-1 >= i and len(code[y])-1 <= j:
                    perm_code_min_max_ord[i][j][x].append(ord(y))

#same as perm_code_min_max_ord but in a np.array
perm_code_min_max_ord_np = np.full((4,4,100,15),-100, dtype=int)

#number of subpatterns in perm_code_min_max_ord[i][j][x]
perm_code_min_max_ord_np_n = np.empty((4,4,100), dtype=int)

for i in range(4):
    for j in range(4):
        for x in perm_code:
            perm_code_min_max_ord_np_n[i,j,ord(x)] = len(perm_code_min_max_ord[i][j][x])
            for k in range(len(perm_code_min_max_ord[i][j][x])):
                perm_code_min_max_ord_np[i,j,ord(x),k] = perm_code_min_max_ord[i][j][x][k]

# minus_set[x][y] is what is left in x after removing y
# minus_set[x][x] is not set
minus_set = {}
for c in complements:
    minus_set[c] = {}
    for x,y in complements[c]:
        minus_set[c][x] = y
        minus_set[c][y] = x


def get_M_U(pattern, contextD, index_mut=0):
    """calculates the number of mutated and unmutated sites that match a IUPAC pattern

    Args:
        pattern (str): IUPAC pattern
        contextD (dict): The number of mutated and unmutated sites for each k-mer
        index_mut (int, optional): To be used for multi class version. Defaults to 0.

    Returns:
        (M,U): number of mutated (M) and unmutated (U) sites
    """
    M = None
    U = None
    for context in matches(pattern):
        tup = contextD[context]
        nm = tup[index_mut]
        nu = tup[-1]
        if M is None:
            M = nm
            U = nu
        else:
            M += nm
            U += nu
    return M, U


def pattern_level(pattern):
    """caculates the level of a pattern
       all k-mers a level 0
       a pattern that needs to be split x times with complements to have only k-mers left has level x 

    Args:
        pattern (str): IUPAC pattern

    Returns:
        int: the level
    """
    return sum(len(code[x])-1 for x in pattern)

# TODO: rename to index_cardinality_table or similar
def get_genpat_pos_level(genpat):
    return [len(perm_code[x]) for x in genpat]

# TODO: rename to prefix_cardinality_table or similar
def get_cum_genpat_pos_level(genpat):
    gppl = get_genpat_pos_level(genpat)
    L = []
    s = 1
    for x in gppl:
        L.append(s)
        s *= x
    return L


class PatternEnumeration():
    def __init__(self, general_pattern):
        self.genpat = general_pattern
        self.gppl = get_genpat_pos_level(general_pattern)
        self.cgppl = get_cum_genpat_pos_level(general_pattern)
    
    def pattern2num(self, pattern):
        s = 0
        for i in range(len(self.genpat)):
            s += perm_code_no[self.genpat[i]][pattern[i]] * self.cgppl[i]
        return s

    def num2pattern(self, num):
        num=int(num)
        pat = ''
        for i in range(len(self.genpat)):
            k = num % self.gppl[i]
            num = num // self.gppl[i]
            pat = pat + perm_code[self.genpat[i]][k]
        return pat


def pattern2num_new_ord(cgppl, genpat, pat):
    s = 0
    for i in range(len(genpat)):
        s += perm_code_no_np[genpat[i]][pat[i]] * cgppl[i]
    return s

def LCA_pattern_of_kmers(contexts):
    '''Least common ancestor (LCA) pattern of a list of k-mers'''
    pat_L = []
    for i in range(len(contexts[0])):
        s = frozenset(x[i] for x in contexts)
        pat_L.append(inv_code[s])
    return ''.join(pat_L)

def LCA_pattern_of_patterns(patterns):
    '''Least common ancestor (LCA) pattern of a list of patterns'''
    pat_L = []
    for i in range(len(patterns[0])):
        s = frozenset(list(chain(*[code[x[i]] for x in patterns])))
        pat_L.append(inv_code[s])
    return ''.join(pat_L)


def match(pattern, context):
    """check if a k-mer matches a pattern

    Args:
        pattern (str): IUPAC pattern
        context (str): k-mer

    Returns:
        bool: True if context matches pattern
    """
    for (p, c) in zip(pattern, context):
        if c not in code[p]:
            return False
    return True

def matches(pattern):
    """generate all k-mers that match a pattern

    Args:
        pattern (str): IUPAC pattern

    Yields:
        str: k-mer
    """
    if len(pattern) == 0:
        yield ''
    else:
        for y in matches(pattern[1:]):
            for x in code[pattern[0]]:
                yield x+y

def subpatterns_level(pattern, level):
    cur_level = pattern_level(pattern)
    rest_level = cur_level - (len(code[pattern[0]])-1)
    for x in perm_code_min_max[max(0,level - rest_level)][min(level,3)][pattern[0]]:#perm_code[pattern[0]]:
        new_level = cur_level - (len(code[pattern[0]]) - len(code[x]))
        if len(pattern)>1 and (new_level >= level):
            for y in subpatterns_level(pattern[1:], level-(len(code[x])-1)):
                yield x+y
        elif len(pattern)==1 and new_level == level:
            yield x

def subpatterns_level_ord(pattern, level):
    cur_level = pattern_level(pattern)
    rest_level = cur_level - (len(code[pattern[0]])-1)
    for x in perm_code_min_max[max(0,level - rest_level)][min(level,3)][pattern[0]]:#perm_code[pattern[0]]:
        new_level = cur_level - (len(code[pattern[0]]) - len(code[x]))
        if len(pattern)>1 and (new_level >= level):
            for y in subpatterns_level_ord(pattern[1:], level-(len(code[x])-1)):
                yield [ord(x)]+y
        elif len(pattern)==1 and new_level == level:
            yield [ord(x)]

code_lev = {}
code_lev_ord = {}
for x in code:
    code_lev[x] = len(code[x]) -1
    code_lev_ord[ord(x)] = len(code[x]) -1

def subpatterns_level_ord_new(pattern, cur_level, level):
    rest_level = cur_level - code_lev[pattern[0]]
    for x in perm_code_min_max_ord[max(0,level - rest_level)][min(level,3)][pattern[0]]:#perm_code[pattern[0]]:
        new_level = cur_level - (code_lev[pattern[0]] - code_lev_ord[x])
        if len(pattern)>1 and (new_level >= level):
            for y in subpatterns_level_ord_new(pattern[1:],cur_level-code_lev[pattern[0]], level-code_lev_ord[x]):
                yield (x,)+y
        elif len(pattern)==1 and new_level == level:
            yield (x,)

code_lev_ord_np = np.full(90,-100,dtype=int)
for x in code_lev_ord:
    code_lev_ord_np[x] = code_lev_ord[x]

@njit
def subpatterns_level_ord_np(pattern, cur_level, level):
    """Fast solution to generate all subpattern at a specific level

    Args:
        pattern: general_pattern encoded as tuple of ord(x) where x is a IUPAC character
        cur_level: level of general pattern
        level ([type]): [description]

    Yields:
        patterns encoded as tuples of ord(x) where x is a IUPAC character
    """
    rest_level = cur_level - code_lev_ord_np[pattern[0]]
    imin = max(0,level - rest_level)
    imax = min(level,3)
    for i in range(perm_code_min_max_ord_np_n[imin,imax,pattern[0]]):
        x = perm_code_min_max_ord_np[imin,imax,pattern[0],i]
        new_level = cur_level - (code_lev_ord_np[pattern[0]] - code_lev_ord_np[x])
        if len(pattern)>1 and (new_level >= level):
            for y in subpatterns_level_ord_np(pattern[1:],cur_level-code_lev_ord_np[pattern[0]], level-code_lev_ord_np[x]):
                yield (x,)+y
        elif len(pattern)==1 and new_level == level:
            yield (x,)


def subpatterns(pattern):
    """Generate all subpatterns of pattern

    Args:
        pattern (str): IUPAC pattern

    Yields:
        str: IUPAC subpattern of pattern
    """
    if len(pattern) == 0:
        yield ''
    else:
        for y in subpatterns(pattern[1:]):
            for x in perm_code[pattern[0]]:
                yield x+y


def generality(pat):
    """Number of kmers that match a given pattern

    Args:
        pat (str): IUPAC pattern

    Returns:
        int: Number of kmers that match pat
    """
    res = 1
    for x in pat:
        res *= len(code[x])
    return res

def pattern_max(general_pattern):
    """Number of subpatterns of a pattern

    Args:
        general_pattern (str): IUPAC pattern

    Returns:
        int: Number of subpatterns of general_pattern
    """
    res = 1
    for i in range(len(general_pattern)):
        res *= len(perm_code[general_pattern[i]])
    return res









# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kmerpapa', 'kmerpapa.algorithms']

package_data = \
{'': ['*']}

install_requires = \
['numba>=0.54.1,<0.55.0', 'numpy>=1.17,<1.21', 'scipy>=1.7.1,<2.0.0']

entry_points = \
{'console_scripts': ['kmerpapa = kmerpapa.cli:main']}

setup_kwargs = {
    'name': 'kmerpapa',
    'version': '0.2.1',
    'description': 'Tool to calculate a k-mer pattern partition from position specific k-mer counts.',
    'long_description': '# kmerPaPa\nTool to calculate a "k-mer pattern partition" from position specific k-mer counts. This can for instance be used to train a mutation rate model.\n\n## Requirements\nkmerPaPa requires Python 3.7 or above.\n\n## Installation\nkmerPaPa can be installed using pip:\n```\npip install kmerpapa\n```\nor using [pipx](https://pypa.github.io/pipx/):\n```\npipx install kmerpapa\n```\n\n## Test data\nThe test data files used in the usage examples below can be downloaded from the test_data directory in the project\'s github repository:\n```\nwget https://github.com/BesenbacherLab/kmerPaPa/raw/main/test_data/mutated_5mers.txt\nwget https://github.com/BesenbacherLab/kmerPaPa/raw/main/test_data/background_5mers.txt\n```\n\n## Usage\nIf we want to train a mutation rate model then the input data should specifiy the number of times each k-mer is observed mutated and unmutated. One option is to have one file with the mutated k-mer counts (positive) and one file with the count of k-mers in the whole genome (background).  We can then run kmerpapa like this:\n```\nkmerpapa --positive mutated_5mers.txt \\\n         --background background_5mers.txt \\\n         --penalty_values 3 5 7\n```\nThe above command will first use cross validation to find the best penalty value between the values 3,5 and 7. Then it will find the optimal k-mer patter partiton using that penalty value.\nIf both a list of penalty values and a list of pseudo-counts are specified then all combinations of values will be tested during cross validation:\n```\nkmerpapa --positive mutated_5mers.txt \\\n         --background background_5mers.txt \\\n         --penalty_values 3 5 6 \\\n         --pseudo_counts 0.5 1 10\n```\nIf only a single combination of penalty_value and pseudo_count is provided then the default is not to run cross validation unless "--n_folds" option or the "CV_only" is used. The "CV_only" option can be used together with "--CVfile" option to parallelize grid search.\nFx. using bash:\n```\nfor c in 3 5 6; do\n    for a in 0.5 1 10; do\n        kmerpapa --positive mutated_5mers.txt \\\n         --background background_5mers.txt \\\n         --penalty_values $c \\\n         --pseudo_counts $a \\\n         --CV_only --CVfile CV_results_c${c}_a${a}.txt &\n    done\ndone\n```\n',
    'author': 'Søren Besenbacher',
    'author_email': 'besenbacher@clin.au.dk',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/besenbacherLab/kmerpapa',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



import numpy as np
from kmerpapa.pattern_utils import matches, PatternEnumeration


def sample(m, colors, itype, prng):
    """ Sample from multivariate hypergeometric distribution
    colors should be n_kmers*2 array (unmutated and mutated are seperate colors)
    # from: https://stackoverflow.com/questions/35734026/numpy-drawing-from-urn
    Args:
        m : number balls to draw from the urn
        colors : one-dimensional array of number balls of each color in the urn
        itype : integer type should be either np.uint32 or np.uint64
        prng : numpy random number generator

    Returns:
        One-dimensional array with the same length as `colors` containing the
        number of balls of each color in a random sample.
    """
    remaining = np.cumsum(colors[::-1])[::-1]
    result = np.zeros(len(colors), dtype=itype)
    for i in range(len(colors)-1):
        if m < 1:
            break
        result[i] = prng.hypergeometric(colors[i], remaining[i+1], m)
        m -= result[i]
    result[-1] = m
    return result


def make_all_folds_contextD_patterns(contextD, U_mem, M_mem, general_pattern, prng, itype = np.uint64):
    """Divide counts from contextD into U_mem and M_mem by sampling from multivariate hypergeometric distribution

    Args:
        contextD ({kmer:(n_mut, n_unmut)}): Dictionary with counts for kmer
        U_mem ((n_folds x n_patterns) np.array): Array to be filled with unmutated counts for each kmer for each fold
        M_mem ((n_folds x n_pattern) np.array): Array to be filled with mutated counts for each kmer for each fold
        general_pattern (str): The general pattern. fx "NNN".
        prng : numpy random number generator
        itype : integer type should be either np.uint32 or np.uint64
    """
    PE = PatternEnumeration(general_pattern)
    contexts = list(contextD.keys())
    contexts.sort()
    colors = np.empty(2*len(contexts), dtype=itype)
    for i in range(len(contexts)):
        nm, nu = contextD[contexts[i]]
        colors[i]  = nm
        colors[len(contexts)+i] = nu
    n = colors.sum()
    n_folds = U_mem.shape[1]
    n_samples = n//n_folds
    samples = np.empty((2*len(contexts), n_folds), dtype=itype)
    for i in range(n_folds-1):
        s = sample(n_samples, colors, itype, prng)
        samples[:,i] = s
        colors -= s
    samples[:,n_folds-1] = colors
    for i in range(len(contexts)):
        context = contexts[i]
        pat_num = PE.pattern2num(context)
        M_mem[pat_num] = samples[i]
        U_mem[pat_num] = samples[i+len(contexts)]


def make_all_folds_contextD_kmers(contextD, U_mem, M_mem, general_pattern, prng):
    """Divide counts from contextD into U_mem and M_mem by sampling from multivariate hypergeometric distribution

    Args:
        contextD ({kmer:(n_mut, n_unmut)}): Dictionary with counts for kmer
        U_mem ((n_folds x n_kmers) np.array): Array to be filled with unmutated counts for each kmer for each fold
        M_mem ((n_folds x n_kmers) np.array): Array to be filled with mutated counts for each kmer for each fold
        general_pattern (str): The general pattern. fx "NNN".
        prng : numpy random number generator
    """
    contexts = list(matches(general_pattern))
    itype = np.uint64
    colors = np.zeros(2*len(contexts), dtype=np.uint64)
    for i in range(len(contexts)):
        nm, nu = contextD[contexts[i]]
        colors[i]  = nm
        colors[len(contexts)+i] = nu
    n = colors.sum()
    n_folds = U_mem.shape[1]
    n_samples = n//n_folds
    samples = np.zeros((2*len(contexts), n_folds), dtype=np.uint64)
    #Fill the first n_folds-1 folds:
    for i in range(n_folds-1):
        s = sample(n_samples, colors, itype, prng)
        samples[:,i] = s
        colors -= s
    #Fill the last fold with the remaining counts:
    samples[:,n_folds-1] = colors
    for i in range(len(contexts)):
        M_mem[i] = samples[i]
        U_mem[i] = samples[i+len(contexts)]
    

def make_all_folds(n_mut, n_unmut, U_mem, M_mem):
    """Divide counts from n_mut and n_unmut into U_mem and M_mem by sampling from multivariate hypergeometric distribution

    Args:
        n_mut (int np.array): number of mutated sites for each kmer
        n_unmut (int np.array): number of unmutated sites for each kmer
        U_mem ((n_folds x n_kmers) np.array): Array to be filled with unmutated counts for each kmer for each fold
        M_mem ((n_folds x n_kmers) np.array): Array to be filled with mutated counts for each kmer for each fold
        prng: numpy random number generator
    """
    itype = n_mut.dtype
    prng = np.random.RandomState()
    n_context = len(n_mut)
    colors = np.concatenate((n_unmut, n_mut))
    n = n_mut.sum() + n_unmut.sum()
    n_folds = U_mem.shape[1]
    n_samples = n//n_folds
    for i in range(n_folds-1):
        samples = sample(n_samples, colors, itype, prng)
        colors -= samples
        U_mem[:n_context,i] = samples[:n_context]
        M_mem[:n_context,i] = samples[n_context:]
    U_mem[:n_context,n_folds-1] = colors[:n_context]
    M_mem[:n_context,n_folds-1] = colors[n_context:]






from kmerpapa.pattern_utils import code, matches, set_code, set_perm_code, inv_code

class Pattern():
    """Class representing a k-mer pattern
    """
    def __init__(self, pattern_string):
        self.pattern = pattern_string

    def __contains__(self, context):
        '''Is context string covered by the pattern?'''
        for (p, c) in zip(self.pattern, context):
            if c not in code[p]:
                return False
        return True

    def __str__(self):
        return self.pattern

    def __repr__(self):
        return self.pattern

    def __len__(self):
        return len(self.pattern)

    def __iter__(self):
       return matches(self.pattern)

    def __and__(self, other):
        '''intersection between two patterns'''
        res = []
        for (c1, c2) in zip(self.pattern, other.pattern):
            scode = set_code[c1] & set_code[c2]
            if len(scode) == 0:
                return None
            res.append(inv_code[scode])
        return Pattern(''.join(res))

    def __le__(self, other):
        '''Is other pattern a subpattern of this'''
        for (x, y) in zip(self.pattern, other.pattern):
            if x not in set_perm_code[y]:
                return False
        return True

    def cardinality(self):
        '''number of contexts matching pattern'''
        res = 1
        for x in self.pattern:
            res *= len(code[x])
        return res


class PatternPartition():
    """Class representing a k-mer Pattern Partition
    """
    def __init__(self, patterns, superPattern=None, strandSymmetry=True):
        ## validate that it is a partition ##
        patterns.sort()
        self.patterns = [Pattern(p) for p in patterns]
        if superPattern is None:
            sp = 'N' * len(patterns[0])
            if strandSymmetry:
                radius = len(patterns[0])//2
                sp =  'N'*radius + 'M' + 'N'*radius
            self.superPattern = Pattern(sp)
        else:
            self.superPattern = Pattern(superPattern)
        n_matches = 0

        for i in range(len(self.patterns)):
            n_matches += self.patterns[i].cardinality()

            #assert that pattern i is a subpattern of the superPattern
            assert self.patterns[i] <= self.superPattern, \
              "pattern #%r (%r) is not a subpattern of the superPattern (%r)" \
              %(i, self.patterns[i], self.superPattern)

            # Out-commentet because it is slow and most errors should be caught by n_matches assertion
            #for j in range(i+1,len(self.patterns)):
            #    #assert that common set between pattern i and pattern j is the empty set:
            #    assert (self.patterns[i] & self.patterns[j]) is None, \
            #      "the common set between pattern #%r (%r) and pattern #%r (%r) is not empty (%r)" \
            #      %(i, self.patterns[i], j, self.patterns[j], self.patterns[i] & self.patterns[j])

        #assert that the patterns cover the super pattern:
        assert n_matches == self.superPattern.cardinality(), \
          "the patterns does not cover the superPattern (%r)" %(self.superPattern)

    def __len__(self):
        return len(self.patterns)

    def pattern_length(self):
        return len(self.patterns[0])

    def __getitem__(self, context):
        for p in self.patterns:
            if context in p:
                return p
        return None

    def __str__(self):
        res = ["[PatternPartition:"]
        for pattern in self.patterns:
            res.append(str(pattern) + ' ' + str(pattern.cardinality()))
        res.append('-' * len(self.patterns[0]))
        res.append(str(self.superPattern) + ' ' + str(self.superPattern.cardinality()) + ']')
        return '\n '.join(res)
